(* An environment model interpreter for arithmetic expressions with
   integers and addition, let expressions, functions, and application. *)

(**********************************************************)
(* SYNTAX *)
(**********************************************************)

(* the type of identifiers *)
type id = string

(* the type of the AST for the expressions
 * e :: = x | i | e1 + e2 | let x = e1 in e2 | fun x -> e | e1 e2
 *)
type expr =
  | Var of id
  | Int of int
  | Add of expr * expr
  | Let of id * expr * expr
  | Fun of id * expr
  | App of expr * expr

(* the type of environments, which are represented as association lists *)
type env = (id * value) list

(* the type of values, which has to be different from the type of the AST,
   because closures are not expressions. *)
and value =
  | VInt of int
  | VClosure of id * expr * env
    (* [VClosure(x, e, env)] represents the closure (|fun x -> e, env|) *)

(**********************************************************)
(* DYNAMIC SEMANTICS *)
(**********************************************************)

(* three predefined exceptions to use as errors *)
let unbound_variable = Failure "Unbound variable"
let cannot_add = Failure "Cannot add non-integer"
let not_function = Failure "Cannot apply non-function"

(* [bind x v env] is env[x->v]. *)
let bind x v env =
  (x,v)::env

(* [lookup x env] is env(x).  If [x] is not bound in [env],
   [lookup x env] raises [unbound_variable]. *)
let lookup x env =
  match List.assoc_opt x env with
  | None -> raise unbound_variable
  | Some v -> v

(* the initial environment is empty *)
let initial_env = []

(*
The evaluation rules are as follows.

<x, env> ==> v
  if env(x) = v

<i, env> ==> i

<e1 + e2, env> ==> i
  if <e1, env> ==> i1
     <e2, env> ==> i2
     i is i1+i2

<let x = e1 in e2, env> ==> v
  if <e1, env> ==> v1
     <e2, env[x->v1]> ==> v

<fun x -> e, env> ==> (|fun x -> e, env|)

<e1 e2, env> ==> v
  if <e1, env> ==> (|fun x -> e, env_cl|)
     <e2, env> ==> v2
     <e, env_cl[x->v2]> ==> v
*)

(* [eval e env] is [v] iff <e, env> ==> v *)
let rec eval e env = match e with
  | Var x -> lookup x env
  | Int i -> VInt i
  | Add (e1, e2) -> eval_add e1 e2 env
  | Let (x, e1, e2) -> eval_let x e1 e2 env
  | Fun (x, e) -> VClosure (x, e, env)
  | App (e1, e2) -> eval_app e1 e2 env

and eval_add e1 e2 env =
  let v1 = eval e1 env in
  match v1 with
  | VClosure _ -> raise cannot_add
  | VInt i1 -> eval_add_int i1 e2 env

and eval_add_int i1 e2 env =
  let v2 = eval e2 env in
  match v2 with
  | VClosure _ -> raise cannot_add
  | VInt i2 -> VInt (i1+i2)

and eval_let x e1 e2 env =
  let v = eval e1 env in
  eval e2 (bind x v env)

and eval_app e1 e2 env =
  let v1 = eval e1 env in
  match v1 with
  | VInt _ -> raise not_function
  | VClosure (x, e, env_cl) -> eval_app_closure x e env_cl e2 env

and eval_app_closure x e_body env_cl e_arg env_app =
  let v = eval e_arg env_app in
  eval e_body (bind x v env_cl)

let interp e =
  eval e initial_env
